import numpy as np
import time 
from math import sqrt
import matplotlib.pyplot as plt
import copy
import scipy.io as sio


# Plotting function for the CGM part
def plot_result(output, X, iAL = False):
    opt_val = 57.037877018731310


    plt.figure(figsize=(12, 12))
    plt.subplot(221)
    plt.loglog(output['cur_time'], output['feasibility1'])#, 'go--', linewidth=2, markersize=12))
    plt.xlabel('Time (s)',fontsize=15)
    plt.ylabel('$\|X1-1\|/1$',fontsize=15)
    plt.grid(True)
    if not iAL:
        plt.subplot(222)
        plt.loglog(output['cur_time'], output['feasibility2'])
        plt.xlabel('Time (s)',fontsize=15)
        plt.ylabel('dist$(X, \mathbb{R}^{n}_+)$',fontsize=15)
        plt.grid(True)

    plt.subplot(223)
    obj_res = np.reshape(np.abs(np.array(output['objective']) - opt_val)/opt_val, (len(output['objective']),))
    plt.loglog(output['cur_time'], obj_res)
    plt.xlabel('Time (s)',fontsize=15)
    plt.ylabel('$(f(X) - f^*)/f^*$',fontsize=15)
    plt.title('Relative objective residual',fontsize=15)
    plt.grid(True)
    
    plt.subplot(224)
    plt.imshow(X)
    plt.title('SDP solution',fontsize=15)
    plt.show()

    
def plot_compare_result(output1,output2, X1, X2):
    opt_val = 57.037877018731310

    plt.figure(figsize=(12, 12))
    plt.subplot(221)
    plt.loglog(output1['cur_time'], output1['feasibility1'], label='CGM')
    # sqrt(2) division to account for the fact that the constraint is put twice
    plt.loglog(output2['cur_time'], output2['feasibility1']/sqrt(2), label ='iAL') 
    plt.xlabel('Time (s)',fontsize=15)
    plt.ylabel('$\|X1-1\|/1$',fontsize=15)
    plt.title('Feasibility: row-sum to 1')
    plt.grid(True)
    plt.legend()
    
    plt.subplot(222)
    obj_res1 = np.reshape(np.abs(np.array(output1['objective']) - opt_val)/opt_val, (len(output1['objective'])))
    obj_res2 = np.reshape(np.abs(np.array(output2['objective']) - opt_val)/opt_val, (len(output2['objective'])))


    plt.loglog(output1['cur_time'], obj_res1, label='CGM')
    plt.loglog(output2['cur_time'], obj_res2, label ='iAL')

    
    plt.xlabel('Time (s)',fontsize=15)
    plt.ylabel('$(f(X) - f^*)/f^*$',fontsize=15)
    plt.title('Relative objective residual',fontsize=15)
    plt.grid(True)
    plt.legend()
    
    plt.subplot(223)
    plt.imshow(X1)
    plt.title('SDP solution from CGM',fontsize=15)
    
    plt.subplot(224)
    plt.imshow(X2)
    plt.title('SDP solution from iAL',fontsize=15)
    plt.show()




def  solve_apgm_clustering(myFx, x0,maxIter,proj,step_size):
    # Function nonconvex accelerated proximal gradient method.
    # This is the black box solver providing the solution of the optimization problem.
    # It performs standard acccelerated proximal gradient method: 
    # Input
    #   myFx:   the function handle of the function to be optimized 
    #   x0:     vector of initial start
    #   maxIter: max number of iteration  
    #   proj: the projection operator
    #   step_size: the step size of the method
    # Output
    #   x1:   optimized variable
    #   f1:   optimized function value
    #   k:    iteration number
    #   alpha: the final step size

    gradToler = 1e-10 # tolerance for the norm of the slope

    n = len(x0)

    res_period = 50
    
    # acceleration params
    alpha = step_size
    x1 = x0
    z1 = x0

    t0 = 1
    t1 = 1
    gx1 = float('Inf')*np.ones(x0.shape)
    k =1
    while True:
        if k > maxIter:
            break
        if k > 1:
            fnorm = np.linalg.norm(gx1,2)
            if fnorm < gradToler:
                break
        
        y1 = x1 + t0/t1*(z1 - x1) + (t0-1)/t1*(x1 - x0);

        fy1,gy1= myFx(y1)
        z1 = proj(y1 - alpha*gy1)
        
        fx1,gx1=myFx(x1)
        v1 = proj(x1 - alpha*gx1)
        
        t0 = t1
        t1 = (np.sqrt(4*t1**2 + 1) + 1)/2
        x0 = x1
        fz1,_= myFx(z1)
        fv1,_= myFx(v1)
        # 
        if fz1 <= fv1:
            x1 = z1
        else:
            x1 = v1

        k = k + 1

    k = k-1


    return x1,fx1,k,alpha

def update_params_nonadaptive(feas_cur, iter, sigma_0, beta_0):
    # This function computes the parameters for the new iteration depending on the predefined adaptive rule
    # INPUTS:  feas_cur: current feasibility value
    #          iter    : current iteration number
    #          sigma_0 : initial value for the dual step size
    #          beta_0  : initial value for the penalty parameter
    # OUTPUTS: beta_   : updated penalty parameter
    #          sigma_  : updated dual step size
    # 
    # Author: Mehmet Fatih Sahin
    # Date:   12.11.2018
    if iter < 1:
        beta_ = beta_0;
        sigma_ = sigma_0;
        return beta_, sigma_

    beta_ = beta_0*sqrt(iter);
    sigma_ = min(sigma_0/sqrt(iter),10*sigma_0/(iter*feas_cur*np.log(iter+1)**2+np.finfo(float).eps));
    return beta_, sigma_

def init_hist_track(params):
    # Initialization of the parameters, returns the output dict.
    output = dict()
    output['cur_time'] = np.nan*np.zeros([params['iters'], 1]);
    output['feasibility1']  = np.nan*np.zeros([params['iters'], 1]);
    output['objective']   = np.nan*np.zeros([params['iters'], 1]);
    output['grad_mapping'] = np.nan*np.zeros([params['iters'], 1]);
    output['sigma']   = np.nan*np.zeros([params['iters'], 1]);
    output['beta']   = np.nan*np.zeros([params['iters'], 1]);
    output['gamma']   = np.nan*np.zeros([params['iters'], 1]);
    output['norm_dual'] = np.nan*np.zeros([params['iters'], 1]);
    output['lagr']   = np.nan*np.zeros([params['iters'], 1]);
    output['dist_'] = np.nan*np.zeros([params['iters'], 1]);
    output['iter'] = np.nan*np.zeros([params['iters'], 1]);
    return output
    
def init_iters(params):
    # initialize the iterations and the parameters 

    Ucur = params['Uinit']
    Ucur = Ucur/np.linalg.norm(Ucur, 'fro');
    
    yc = params['ycinit']; # center point to be updated
    ysum = yc
    Usum = Ucur
    gamma_ = params['gamma']

    beta_, sigma_ = params['beta0'], params['sigma0']
    beta_0 = beta_

    return Ucur, yc, ysum, Usum, gamma_, beta_, beta_0, sigma_
    
        

def track_hist(ii, output, start, Ucur, Uold, beta_, sigma_, gamma_, yc, get_feas, gradFunc, tot_iter, f_u):
    # Save the result of the current iteration.
    
    output['cur_time'][ii] = time.time()-start
    output['feasibility1'][ii]  = np.linalg.norm(get_feas(Ucur),2)/Ucur.shape[0]
    output['objective'][ii]   = f_u(Ucur)
    output['grad_mapping'][ii] = 1/gamma_**2 * np.linalg.norm(Ucur-Uold, 2).item()**2
    current_obj = output['objective'][ii]
    output['sigma'][ii] = sigma_
    output['gamma'][ii]   = gamma_
    output['norm_dual'][ii] = np.linalg.norm(yc,2);
    output['beta'][ii]  = beta_;
    output['lagr'][ii], _  = gradFunc(Uold);
    output['iter'][ii] = tot_iter;
    return output

def verbosity(ii, params, output,tot_iter):
    # Plot the results of the iteration
        if (params['verbose']) and ((ii % params['verbose'])==0 or ii == 0):
            if (ii%20*params['verbose'])==0:
                print('    Iter.      |    Feas        |      Objective         |       Grad Mapping        |     Nfeval');
                print('---------------------------------------------------------------------------------------------------');
            print('{:>7d}             {:>6.10f}        {:>10.5f}        {:>15.3f}             {:>5d}'.format(
                ii, output['feasibility1'][ii].item(), output['objective'][ii].item(), output['grad_mapping'][ii].item(), tot_iter))
            
            
## FUNCTIONS FOR THE SDP ROUNDING

def remap_centers(assign,k):
    class_vec = np.zeros([10,10])
    max_class = np.zeros([10,2])
    remap_vec = -1*np.ones([10])
    
    # Computing the number of assignments per class (the classes go 0-99, 100-199,...) 
    for l in range(10):
        class_loc = assign[100*(l):100*(l+1)-1]
        for i in range(10):
            class_vec[l,i] = sum(class_loc==i).item()
        max_class[l] = [np.max(class_vec[l,:]), np.argmax(class_vec[l,:])]

    # Remapping the cluster with the largest number of elements to the actual one iteratively
    # Plotting the evolution of class_vec, pos_remap and max_class helps to understand it 
    # easily ;)
    for l in range(10):  
        pos_remap = np.argmax(max_class[:,0])
        remap_vec[int(max_class[pos_remap,1])] = pos_remap
        class_vec[:,int(max_class[pos_remap,1])] = 0
        class_vec[pos_remap,:] = 0
        for i in range(10):
            max_class[i] = [np.max(class_vec[i,:]), np.argmax(class_vec[i,:])]

    return remap_vec




def sdp_rounding(X, k):
    digits  = sio.loadmat('data/fashion_mnist_digits.mat')['digits']
    X = digits.dot(X)
    
    N=X.shape[1]
    # computation of an affinity matrix identifying repeated denoised points
    affinity=np.zeros([N,N]);
    for i in range(N): 
        for j in range(N):
            if np.linalg.norm(X[:,i]-X[:,j])<1e-3:
                affinity[i,j]=1;
                affinity[j,i]=1;
    # centers are k most popular points
    centers=np.zeros([k,k])
    for t in range(k):
        s = np.sum(affinity,0)
        idx = np.argmax(s)

        aux = copy.deepcopy(affinity[:,idx])
        centers[t,:]=X[:, idx]
        for i in range(N): 
            if aux[i]==1:
                affinity[i,:]=0
                affinity[:,i]=0
        
    # assignment of points to closest center
    ind=np.zeros([N,1]);
    for i in range(N):
        aux=np.zeros([k,1]);
        for t in range(k):
            aux[t,0]=np.linalg.norm(X[:,i].T - centers[t,:],2);
        ind[i,0]= np.argmin(aux);
    assignment=ind
    
    # remapping to the correct clusters, i.e. first cluster should be 0, ...
    assignment_remap = np.zeros([N,1]);
    remap_vec = remap_centers(assignment,k)
    centers_remap =np.zeros([k,k]); 
    for i in range(N):
            assignment_remap[i] = remap_vec[int(assignment[i])]
    for loc, map_ in enumerate(remap_vec):
        centers_remap[loc,:] = centers[int(map_),:]
    
    return centers_remap, assignment_remap

def misclassification_rate(assignment):
    labels = sio.loadmat('data/fashion_mnist_labels.mat')['labels']-1
    return np.sum(assignment!=labels)/len(assignment)

def vis_samples(assignment):
    images = sio.loadmat('data/fashion_mnist_images.mat')['images']
    assignment=assignment.astype(int)
    classes = ['T-shirt/top','Trouser','Pullover','Dress','Coat','Sandal','Shirt','Sneaker','Bag','Ankle boot']
    labels = sio.loadmat('data/fashion_mnist_labels.mat')['labels']-1
    rand_samp = np.random.randint(0,1000,25)
    plt.figure(figsize=(7,7))
    for i,samp in enumerate(rand_samp):
        plt.subplot(5,5,i+1)
        plt.imshow(1-np.reshape(images[samp],[28,28]), cmap=plt.cm.gray)
        plt.title('Pred. {0}\n Orig. {1}'.format(classes[assignment[samp].item()],classes[labels[samp].item()]))
        plt.axis('off')
    plt.tight_layout()
    plt.show()
    